#include "freelist_internal.h"





/****************************************************************************/
atom_t freelist_new( void **freelist_state, atom_t number_elements, int (*user_data_init_function)(void **user_data, void *user_state), void *user_state )
{
  struct freelist_state
    *fs;

  atom_t
    count = 0;

  assert( freelist_state != NULL );
  // number_elements can be any value in its range
  // user_data_init_function can be NULL

  fs = *freelist_state = abstraction_aligned_malloc( sizeof(struct freelist_state), 16 );

  if( fs != NULL )
  {
    fs->top[FREELIST_POINTER] = NULL;
    fs->top[FREELIST_COUNTER] = 0;
    fs->user_data_init_function = user_data_init_function;
    fs->user_state = user_state;
    fs->element_count = 0;

    if( number_elements > 0 )
      fs->element_count = count = freelist_new_elements( *freelist_state, number_elements );
  }

  return( count );
}





/****************************************************************************/
atom_t freelist_new_elements( void *freelist_state, atom_t number_elements )
{
  struct freelist_element
    *fe;

  atom_t
    loop,
    count = 0;

  assert( freelist_state != NULL );
  assert( number_elements >= 1 );
  // user_data_init_function can be NULL

  for( loop = 0 ; loop < number_elements ; loop++ )
    if( freelist_internal_new_element(freelist_state, (void **) &fe) )
    {
      freelist_push( freelist_state, fe );
      count++;
    }

  return( count );
}





/****************************************************************************/
atom_t freelist_internal_new_element( struct freelist_state *fs, void **freelist_element )
{
  struct freelist_element
    *fe;

  atom_t
    rv = 0;

  assert( fs != NULL );
  assert( freelist_element != NULL );
  // user_data_init_function can be NULL

  /* TRD : basically, does what you'd expect;

           allocates an element
           calls the user init function
           if anything fails, cleans up,
           sets *freelist_element to NULL
           and returns 0
  */

  *freelist_element = fe = abstraction_aligned_malloc( sizeof(struct freelist_element), 16 );

  if( *freelist_element != NULL )
  {
    if( fs->user_data_init_function == NULL )
    {
      fe->user_data = NULL;
      rv = 1;
    }

    if( fs->user_data_init_function != NULL )
    {
      rv = fs->user_data_init_function( &fe->user_data, fs->user_state );

      if( rv == 0 )
      {
        abstraction_aligned_free( *freelist_element );
        *freelist_element = NULL;
      }
    }
  }

  /* TRD : the new element function calls are thread safe
           we need to use an interlockedincrement on the element_count
           otherwise we can get into the race condition on the ++
           (remember - the assignment operator is not atomic, so the
            ++ operator, which is basically a = a + 1, is not thread
            safe and the results of an addition in one thread can
            be overwritten by the result from another)
           and end up with a total count lower than the actual
           total count
  */

  if( rv == 1 )
    abstraction_increment( (atom_t *) &fs->element_count );

  return( rv );
}

