#include "freelist_internal.h"





/****************************************************************************/
void *freelist_pop( void *freelist_state, void **freelist_element )
{
  struct freelist_state
    *fs;

  ALIGN(ALIGN_DOUBLE_POINTER) volatile struct freelist_element
    *fe[FREELIST_PAC_SIZE],
    *fe_next[FREELIST_PAC_SIZE];

  assert( freelist_state != NULL );
  assert( freelist_element != NULL );

  fs = (struct freelist_state *) freelist_state;

  fe[FREELIST_COUNTER] = fs->top[FREELIST_COUNTER];
  fe[FREELIST_POINTER] = fs->top[FREELIST_POINTER];

  do
  {
    if( fe[FREELIST_POINTER] == NULL )
    {
      *freelist_element = NULL;
      return( NULL );
    }

    *freelist_element = (void *) fe[FREELIST_POINTER];

    fe_next[FREELIST_COUNTER] = fe[FREELIST_POINTER]->next[FREELIST_COUNTER];
    fe_next[FREELIST_POINTER] = fe[FREELIST_POINTER]->next[FREELIST_POINTER];
  }
  while( 0 == abstraction_dcas((volatile atom_t *) fs->top, (atom_t *) fe_next, (atom_t *) fe) );

  return( *freelist_element );
}





/****************************************************************************/
void *freelist_guaranteed_pop( void *freelist_state, void **freelist_element )
{
  assert( freelist_state != NULL );
  assert( freelist_element != NULL );

  freelist_internal_new_element( freelist_state, freelist_element );

  return( *freelist_element );
}





/****************************************************************************/
void freelist_push( void *freelist_state, void *freelist_element )
{
  struct freelist_state
    *fs;

  ALIGN(ALIGN_DOUBLE_POINTER) volatile struct freelist_element
    *fe[FREELIST_PAC_SIZE],
    *original_fe_next[FREELIST_PAC_SIZE];

  ALIGN(ALIGN_SINGLE_POINTER) static volatile atom_t
    counter = 0;

  assert( freelist_state != NULL );
  assert( freelist_element != NULL );

  fs = (struct freelist_state *) freelist_state;

  fe[FREELIST_POINTER] = (volatile struct freelist_element *) freelist_element;
  fe[FREELIST_COUNTER] = (volatile struct freelist_element *) abstraction_increment( (atom_t *) &counter );

  original_fe_next[FREELIST_POINTER] = fs->top[FREELIST_POINTER];
  original_fe_next[FREELIST_COUNTER] = fs->top[FREELIST_COUNTER];

  /* TRD : remember, original_fe_next will be updated, non-atomically, to *a* (not *the*) previous
           but more up-to-date version of fs->top by abstraction_dcas(); but also remember, that
           update occurs whether or not a successful swap occurred, so we can't set fe->next
           directly, because after a successful swap, which means fe is now *in* the freelist,
           we could then obliterate its next pointer, since the update of the comparand argument
           is not atomic - someone else may have changed fe->next before we have got round to
           putting our value in it
  */

  do
  {
    fe[FREELIST_POINTER]->next[FREELIST_POINTER] = original_fe_next[FREELIST_POINTER];
    fe[FREELIST_POINTER]->next[FREELIST_COUNTER] = original_fe_next[FREELIST_COUNTER];
  }
  while( 0 == abstraction_dcas((volatile atom_t *) fs->top, (atom_t *) fe, (atom_t *) original_fe_next) );

  return;
}

