#include "queue_internal.h"





/****************************************************************************/
atom_t queue_new( void **queue_state, atom_t number_elements )
{
  struct queue_state
    *qs;

  volatile struct queue_element
    *qe[QUEUE_PAC_SIZE];

  atom_t
    queue_count = 0;

  assert( queue_state != NULL );
  // number_elements can be any value in its range

  qs = *queue_state = abstraction_aligned_malloc( sizeof(struct queue_state), 16 );

  if( qs != NULL )
  {
    // TRD : the size of the freelist is the size of the queue (+1 for the leading dummy element, which is hidden from the caller)
    queue_count = freelist_new( &qs->freelist_state, number_elements+1, queue_internal_freelist_init_function, NULL );

    if( qs->freelist_state != NULL )
    {
      // TRD : we may have allocated the freelist but not a single element
      if( queue_count == 0 )
      {
        freelist_delete( qs->freelist_state, queue_internal_freelist_delete_function, NULL );
        abstraction_aligned_free( *queue_state );
        *queue_state = NULL;
      }

      if( queue_count >= 1 )
      {
        queue_count--;

        queue_internal_new_element_from_freelist( qs, qe, NULL );

        qs->enqueue[QUEUE_POINTER] = qs->dequeue[QUEUE_POINTER] = qe[QUEUE_POINTER];
        qs->enqueue[QUEUE_COUNTER] = qs->dequeue[QUEUE_COUNTER] = qe[QUEUE_COUNTER];
      }
    }

    if( qs->freelist_state == NULL )
    {
      abstraction_aligned_free( *queue_state );
      *queue_state = NULL;
    }
  }

  return( queue_count );
}





/****************************************************************************/
#pragma warning( disable : 4100 )

int queue_internal_freelist_init_function( void **user_data, void *user_state )
{
  int
    rv = 0;

  assert( user_data != NULL );
  assert( user_state == NULL );

  *user_data = abstraction_aligned_malloc( sizeof(struct queue_element), 16 );

  if( *user_data != NULL )
    rv = 1;

  return( rv );
}

#pragma warning( default : 4100 )





/****************************************************************************/
void queue_internal_new_element_from_freelist( struct queue_state *qs, volatile struct queue_element *qe[QUEUE_PAC_SIZE], void *user_data )
{
  void
    *freelist_element;

  assert( qs != NULL );
  assert( qe != NULL );
  // user_data can be any value in its range

  qe[QUEUE_POINTER] = NULL;
  qe[QUEUE_COUNTER] = 0;

  freelist_pop( qs->freelist_state, &freelist_element );

  if( freelist_element != NULL )
    queue_internal_init_element( qe, freelist_element, user_data );

  return;
}





/****************************************************************************/
void queue_internal_guaranteed_new_element_from_freelist( struct queue_state *qs, volatile struct queue_element *qe[QUEUE_PAC_SIZE], void *user_data )
{
  void
    *freelist_element;

  assert( qs != NULL );
  assert( qe != NULL );
  // user_data can be any value in its range

  qe[QUEUE_POINTER] = NULL;
  qe[QUEUE_COUNTER] = 0;

  freelist_guaranteed_pop( qs->freelist_state, &freelist_element );

  if( freelist_element != NULL )
    queue_internal_init_element( qe, freelist_element, user_data );

  return;
}





/****************************************************************************/
void queue_internal_init_element( volatile struct queue_element *qe[QUEUE_PAC_SIZE], void *freelist_element, void *user_data )
{
  ALIGN(ALIGN_SINGLE_POINTER) static volatile atom_t
    counter = 0;

  assert( qe != NULL );
  assert( freelist_element != NULL );
  // user_data can be any value in its range

  freelist_get_user_data_from_element( freelist_element, (void **) &qe[QUEUE_POINTER] );
  qe[QUEUE_COUNTER] = (volatile struct queue_element *) abstraction_increment( (atom_t *) &counter );

  qe[QUEUE_POINTER]->next[QUEUE_POINTER] = NULL;
  qe[QUEUE_POINTER]->next[QUEUE_COUNTER] = 0;

  qe[QUEUE_POINTER]->freelist_element = freelist_element;
  qe[QUEUE_POINTER]->user_data = user_data;

  return;
}

