/***** the library wide include file *****/
#include "liblfds_internal.h"

/***** pragmas *****/

/***** defines *****/
#define FIFO_STATE_UNKNOWN               -1
#define FIFO_STATE_EMPTY                  0
#define FIFO_STATE_ENQUEUE_OUT_OF_PLACE   1
#define FIFO_ATTEMPT_DEQUEUE              2

#define RINGBUFFER_POINTER   0
#define RINGBUFFER_COUNTER   1
#define RINGBUFFER_PAC_SIZE  2

/***** incomplete types *****/
struct ringbuffer_element;

/***** structures *****/
#pragma pack( push, 16 )

struct ringbuffer_state
{
  volatile struct ringbuffer_element
    *enqueue[RINGBUFFER_PAC_SIZE],
    *dequeue[RINGBUFFER_PAC_SIZE];

  void
    *freelist_state;
};

struct ringbuffer_element
{
  volatile struct ringbuffer_element
    *next[RINGBUFFER_PAC_SIZE];

  void
    *freelist_element,
    *user_data;
};

#pragma pack( pop )

struct ringbuffer_internal_init_state
{
  int
    (*user_data_init_function)( void **user_data, void *user_state );

  void
    *user_state;
};

/***** externs *****/

/***** private prototypes *****/
int ringbuffer_internal_enqueue_element( void *ringbuffer_state, struct ringbuffer_element *re );
int ringbuffer_internal_dequeue_element( void *ringbuffer_state, struct ringbuffer_element **re );

int ringbuffer_internal_freelist_init_function( void **user_data, void *user_state );
void ringbuffer_internal_freelist_delete_function( void *user_data, void *user_state );

