#include "stack_internal.h"





/****************************************************************************/
atom_t stack_new( void **stack_state, atom_t number_elements )
{
  struct stack_state
    *ss;

  atom_t
    stack_count = 0;

  assert( stack_state != NULL );
  // number_elements can be any value in its range

  ss = *stack_state = abstraction_aligned_malloc( sizeof(struct stack_state), 16 );

  if( ss != NULL )
  {
    // TRD : the size of the freelist is the size of the stack
    stack_count = freelist_new( &ss->freelist_state, number_elements, stack_internal_freelist_init_function, NULL );

    if( ss->freelist_state == NULL )
    {
      abstraction_aligned_free( *stack_state );
      *stack_state = NULL;
    }

    if( ss->freelist_state != NULL )
    {
      ss->top[STACK_POINTER] = NULL;
      ss->top[STACK_COUNTER] = 0;
    }
  }

  return( stack_count );
}





/****************************************************************************/
#pragma warning( disable : 4100 )

int stack_internal_freelist_init_function( void **user_data, void *user_state )
{
  int
    rv = 0;

  assert( user_data != NULL );
  assert( user_state == NULL );

  *user_data = abstraction_aligned_malloc( sizeof(struct stack_element), 16 );

  if( *user_data != NULL )
    rv = 1;

  return( rv );
}

#pragma warning( default : 4100 )





/****************************************************************************/
void stack_internal_new_element_from_freelist( struct stack_state *ss, volatile struct stack_element *se[STACK_PAC_SIZE], void *user_data )
{
  void
    *freelist_element;

  assert( ss != NULL );
  assert( se != NULL );
  // user_data can be any value in its range

  freelist_pop( ss->freelist_state, &freelist_element );

  if( freelist_element == NULL )
    se[STACK_POINTER] = NULL;

  if( freelist_element != NULL )
    stack_internal_init_element( se, freelist_element, user_data );

  return;
}





/****************************************************************************/
void stack_internal_new_element( struct stack_state *ss, volatile struct stack_element *se[STACK_PAC_SIZE], void *user_data )
{
  void
    *freelist_element;

  assert( ss != NULL );
  assert( se != NULL );
  // user_data can be any value in its range

  freelist_guaranteed_pop( ss->freelist_state, &freelist_element );

  if( freelist_element == NULL )
    se[STACK_POINTER] = NULL;

  if( freelist_element != NULL )
    stack_internal_init_element( se, freelist_element, user_data );

  return;
}





/****************************************************************************/
void stack_internal_init_element( volatile struct stack_element *se[STACK_PAC_SIZE], void *freelist_element, void *user_data )
{
  ALIGN(ALIGN_SINGLE_POINTER) static volatile atom_t
    counter = 0;

  assert( se != NULL );
  assert( freelist_element != NULL );
  // user_data can be any value in its range

  freelist_get_user_data_from_element( freelist_element, (void **) &se[STACK_POINTER] );

  se[STACK_COUNTER] = (volatile struct stack_element *) abstraction_increment( (atom_t *) &counter );

  se[STACK_POINTER]->next[STACK_POINTER] = NULL;
  se[STACK_POINTER]->next[STACK_COUNTER] = 0;

  se[STACK_POINTER]->freelist_element = freelist_element;
  se[STACK_POINTER]->user_data = user_data;

  return;
}

