#include "internal.h"





/****************************************************************************/
void stack_test( void )
{
  unsigned int
    loop,
    cpu_count;

  void
    *stack_state;

  thread_state_t
    *thread_handles;

  /* TRD : there are 5 tests

           1. single reader thread per CPU
              - stack always empty
           2. single writer thread per CPU
              - stack always full
           3. one reader and one writer thread per CPU
              - stack balanced
           4. one reader and two writer threads per CPU
              - stack grows
           5. two reader and one writer thread per CPU
              - stack tends to empty
  */

  cpu_count = abstraction_cpu_count();

  printf( "\n"
          "Stack test\n"
          "==========\n" );

  // TRD : 1. single reader thread per CPU

  printf( "\n"
          "1. single reader thread per CPU\n"
          "===============================\n" );

  stack_new( &stack_state, 10000 );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], stack_internal_thread_reader, stack_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  stack_delete( stack_state, NULL, NULL );

  free( thread_handles );

  // TRD : 2. single writer thread per CPU

  printf( "\n"
          "2. single writer thread per CPU\n"
          "===============================\n" );

  stack_new( &stack_state, 10000 );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 1 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_start( &thread_handles[loop], stack_internal_thread_writer, stack_state );

  for( loop = 0 ; loop < cpu_count ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  stack_delete( stack_state, NULL, NULL );

  free( thread_handles );

  // TRD : 3. one reader and one writer thread per CPU

  printf( "\n"
          "3. one reader and one writer thread per CPU\n"
          "===========================================\n" );

  stack_new( &stack_state, 10000 );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 2 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], stack_internal_thread_reader, stack_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], stack_internal_thread_writer, stack_state );
  }

  for( loop = 0 ; loop < cpu_count * 2 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  stack_delete( stack_state, NULL, NULL );

  free( thread_handles );

  // TRD : 4. one reader and two writer threads per CPU

  printf( "\n"
          "4. one reader and two writer threads per CPU\n"
          "============================================\n" );

  stack_new( &stack_state, 10000 );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], stack_internal_thread_reader, stack_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], stack_internal_thread_writer, stack_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], stack_internal_thread_writer, stack_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  stack_delete( stack_state, NULL, NULL );

  free( thread_handles );

  // TRD : 5. two reader and one writer thread per CPU

  printf( "\n"
          "5. two reader and one writer thread per CPU\n"
          "===========================================\n" );

  stack_new( &stack_state, 10000 );

  thread_handles = malloc( sizeof(thread_state_t) * cpu_count * 3 );

  for( loop = 0 ; loop < cpu_count ; loop++ )
  {
    abstraction_thread_start( &thread_handles[loop], stack_internal_thread_reader, stack_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count], stack_internal_thread_reader, stack_state );
    abstraction_thread_start( &thread_handles[loop+cpu_count*2], stack_internal_thread_writer, stack_state );
  }

  for( loop = 0 ; loop < cpu_count * 3 ; loop++ )
    abstraction_thread_wait( thread_handles[loop] );

  stack_delete( stack_state, NULL, NULL );

  free( thread_handles );

  return;
}





/****************************************************************************/
thread_return_t stack_internal_thread_reader( void *stack_state )
{
  void
    *user_data;

  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( stack_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    if( stack_pop(stack_state, &user_data) )
      count++;
  }

  printf( "read count = %llu\n", count );

  return( EXIT_SUCCESS );
}





/****************************************************************************/
thread_return_t stack_internal_thread_writer( void *stack_state )
{
  time_t
    start_time;

  unsigned long long int
    count = 0;

  assert( stack_state != NULL );

  time( &start_time );

  while( time(NULL) < start_time + 10 )
  {
    // TRD : we don't store any user data
    if( stack_push(stack_state, NULL) )
      count++;
  }

  printf( "write count = %llu\n", count );

  return( EXIT_SUCCESS );
}

